"use strict";

const express = require('express');
const CodeModule = require('../helper/module');
const Jobs = require('../helper/job');
const Db = require('../helper/db');
const Abort = require('../helper/abort');

const Validator = require('baqend/lib/util/Validator');

// overwrite to compile the validation code in a sandbox
Validator.compile = function(managedType, code) {
  const script = CodeModule.compileScript('/code/' + managedType.name + '/validate', code);

  return function onValidate(validators) {
    script.runInNewContext(validators);
  }
};

const codeType = function(bucket, type) {
  return '/code/' + bucket + (type == 'module' ? '' : '/' + type);
};

const callMethod = function(type, req, res) {
  if (!Db.isInitialized()) {
    //signal the orestes server that we are not initialized
    res.status(503).send();
    return Promise.resolve();
  }

  const db = Db.createUserDBForRequest(req);

  const module = CodeModule.require(codeType(req.params.bucket, 'module'), {});
  if (module[type]) {
    return new Promise(function(resolve) {
      resolve(module[type](db, req, res));
    }).then(function() {
      if (!res.headersSent) {
        throw new Abort("The response was not completed with res.send().");
      }
    });
  } else if (module.call) {
    return new Promise(function(resolve) {
      const data = req.method == "GET" ? req.query : req.body;
      resolve(module.call(db, data, req));
    }).then(function(result) {
      if (result) {
        res.status(200).send(result);
      } else {
        res.status(204).send();
      }
    });
  } else {
    res.status(404).send();
    return Promise.resolve();
  }
};

const callHandler = function(req, res) {
  if (!Db.isInitialized()) {
    //signal the orestes server that we are not initialized
    res.status(503).send();
    return Promise.resolve();
  }

  const db = Db.createUserDBForRequest(req);

  let obj;
  const name = req.params.bucket;
  const type = req.params.type;
  const module = CodeModule.require(codeType(req.params.bucket, req.params.type), {});
  const fn = module['on' + type.substring(0, 1).toUpperCase() + type.substring(1)];

  obj = db[name].newInstance();
  obj._metadata.setJson(req.body);
  obj._metadata.setPersistent();

  return new Promise(function(resolve) {
    const validation = obj.validate();
    if (!validation.isValid) {
      throw new Abort('Object is not valid', validation.toJSON());
    } else {
      resolve(fn ? fn.call(obj, db, obj, req) : null);
    }
  }).then(function() {
    //serialize json first to handle collection dirty checking correctly
    if (name === 'jobs.Definition') {
      Jobs.handleJob(obj, type);
    }
    const json = obj.toJSON();
    if (obj._metadata.isDirty && type != 'delete') {
      res.status(200).send(json);
    } else {
      res.status(204).send();
    }
  });
};

const router = express.Router();
router.put("/schema", function(req, res) {
  //location will be an url or just the public port of the orestes server
  Db.initialize(req);
  res.status(204).send();
});

router.put("/code", function(req, res) {
  const codes = req.body;

  CodeModule.clear();

  for (let key in codes) {
    const parts = key.split('/');
    if (parts[1] == 'code') {
      try {
        CodeModule.register(codeType(parts[2], parts[3]), codes[key]);
      } catch (err) {
        Db.logError(err);
      }
    }
  }

  //Give the server enough time to initialize
  Jobs.maintenance(true);

  res.status(204).send();
});

router.put("/code/:bucket/:type", function(req, res) {
  res.set({
    'Content-Type': 'text/plain;charset=utf-8'
  });

  try {
    CodeModule.register(codeType(req.params.bucket, req.params.type), req.body);
  } catch (err) {
    console.error("Invalid Code encountered", err);
  }

  res.status(200).send(req.body);
});

router.delete("/code/:bucket/:type", function(req, res) {
  CodeModule.unregister(codeType(req.params.bucket, req.params.type));

  res.status(204).send();
});

router.post("/code/:bucket/:type", function(req, res, next) {
  callHandler(req, res).catch(next);
});

router.get("/code/:bucket", function(req, res, next) {
  callMethod('get', req, res).catch(next);
});

router.post("/code/:bucket", function(req, res, next) {
  callMethod('post', req, res).catch(next);
});

module.exports = router;
