//Imports
var ArrayList = Java.type("java.util.ArrayList");
var Boolean = Java.type("java.lang.Boolean");
var HashMap = Java.type("java.util.HashMap");
var HashSet = Java.type("java.util.HashSet");
var Map = Java.type("java.util.Map");
var ObjectUpdate = Java.type("info.orestes.events.ObjectUpdate");
var Operation = Java.type("info.orestes.events.Operation");

//global variables
/**
 *  an array like [ "id" ] that holds information on how to extract an object ID from a given JSON document.
 *
 *  copied from callbaq.util.Config#IDFieldNames JavaDoc:
 *
 * The keys by which an object's ID can be retrieved from its corresponding JSON representation. The following
 * example shows an Orestes JSON document and how the object ID could be retrieved in JavaScript:
 * <pre><code>{
     * "key 1": "value 1",
     * "key 2": "value 2",
     * // meta data below
     * "id": "d89eefe1-7e17-4d7f-89dc-7e31317721db",
     * "version": 1,
     * "acl": {}
     * }</code></pre>
 * <p>
 * <pre><code>String[] IDFieldNames = new String[]{ "id" }</code></pre>
 * <p>
 * JavaScript code:
 * <pre><code>var ID = objectJSON["id"]</code></pre>
 * <p>
 * When the ID is nested,  {@link #IDFieldNames} is a multi-valued array:
 * <p>
 * <pre><code>{
     * "key 1": "value 1",
     * "key 2": "value 2",
     * // meta data below
     * "_objectInfo": {
     * "bucket": "Todo",
     * "id": "9c1d7f79-9533-4212-9108-890ad24a1002",
     * "version": 1,
     * "acl": {}
     * }
     * }</code></pre>
 * <p>
 * <pre><code>String[] IDFieldNames = new String[]{ "_objectInfo", "id" }</code></pre>
 * JavaScript code:
 * <pre><code>var ID = objectJSON["_objectInfo"]["id"]</code></pre>
 *
 *
 * @type {Array}
 */
var oidFieldNames = [];
/**
 * other queries to become siftable, an evaluation call is required. To save performance, the corresponding results are cached here.
 *
 * @type {{}}
 */
var queries_siftable = {};
/**
 * a map by which the change operations in the current batch can be retrieved by their respective after-images' JSON objects
 * @type {{}}
 */
var changesByUpdate = {};


/**
 * matches all known complex queries against the given changes and returns the matches.
 *
 * @param changes {List<ObjectUpdate>} a java.util.List of change operations
 * @param relevantQueries {Collection<StreamingQuery>} a subset of all registered queries for which to evaluate the given change
 * operation
 *                (<code>null</code> indicating tha  all registered queries should be evaluated)
 * @returns {Map<ObjectUpdate, Set<StreamingQuery>>} map: changes --> queries --> matches
 */
function multisift(changes, relevantQueries) {
  var matches = new HashMap();

  // initially, parse all JSON documents
  var changesAsJSON = parseChanges(changes);

  for (var i = queries.iterator(); i.hasNext();) {
    var query = i.next();
    if (relevantQueries === null || relevantQueries.contains(query)) {
      // get a siftable form of the query
      var siftQuery = getSiftable(query);
      // assemble JSON documents
      var siftObjects = objectsFromChanges(query, changes, changesAsJSON);
      // Evaluate queries
      var result = sift(siftQuery, siftObjects);
      // store matches
      storeMatches(matches, result, query);
    }
  }

  return matches;
}


/**
 * transforms the given query in a way that it can be used as query parameter for the sift function.
 * Once created, the siftable query is cached in queries_siftable for increased performance.
 *
 * NOTE: at the moment, siftable queries are created using eval --> DANGER!
 *
 * @param query a query
 * @returns {*} a query that can be used in sift.js
 */
function getSiftable(query) {
  var siftQuery = queries_siftable[query.getId()];
  if (typeof siftQuery === 'undefined') {
    siftQuery = eval('(' + query.getQuery() + ')');
    queries_siftable[query.getId()] = siftQuery;
  }
  return siftQuery;
}


/**
 * parses the JSON representations of the incoming change operations and returns the result as a map.
 * @param changes the current batch of change operations
 * @returns {{}} a map from JSON string to parsed JavaScript object
 */
function parseChanges(changes) {
  var changesAsJSON = {};

  for (var listIterator = changes.iterator(); listIterator.hasNext();) {
    var change = listIterator.next();
    var json = JSON.parse(change.getObject().getJson());
    changesAsJSON[change.getObject().getJson()] = json;
    changesByUpdate[JSON.stringify(json)] = change;
  }
  return changesAsJSON;
}

/**
 * @param fieldName a string
 */
function pushIDFieldName(fieldName) {
  oidFieldNames.push(fieldName);
}


/**
 * removes all change operations from the set of given changes that do not match the given query with respect to bucket and operation type.
 * @param query a query
 * @param changes a set of changes
 * @param changesAsJSON a map to retrieve the parsed JavaScript objects given their respective object IDs
 * @returns {Array} all changes that match the query with respect to bucket and operation
 */
function objectsFromChanges(query, changes, changesAsJSON) {
  var result = [];

  for (var listIterator = changes.iterator(); listIterator.hasNext();) {
    var change = listIterator.next();
    if (query.needsProcessing(change)) {
      var json = changesAsJSON[change.getObject().getJson()];
      result.push(json);
    }
  }
  return result;
}


/**
 * returns the object ID of the given object.
 * @param object a JSON document
 * @returns {java.lang.String} the ID of the given object
 */
function getOID(object) {
  var oid = object;
  for (var j in oidFieldNames) {
    oid = oid[oidFieldNames[j]];
  }
  return oid;
}


/**
 * associates the given query with the change operations corresponding to the JSON documents in the given siftResult.
 * @param matches all matches
 * @param siftResult the result of sifting
 * @param query a query
 * @returns all matches of the given query
 */
function storeMatches(matches, siftResult, query) {
  var match;

  for (var i in siftResult) {
    var json = siftResult[i];
    var change = changesByUpdate[JSON.stringify(json)];

    if (matches.get(change) === null) {
      matches.put(change, new HashSet());
    }
    matches.get(change).add(query);
  }

  return matches;
}